#utility functions for demo
import os
import yaml
import sys
import json
import pandas as pd
import numpy as np
import urllib
import xmltodict
import csv
import warnings
import tensorflow as tf

from tqdm import tqdm
from keras_retinanet import models
from keras.utils import multi_gpu_model
from deepforest import _ROOT


def label_to_name(class_dict, label):
    """ Map label to name.
        """
    name = class_dict[label]
    return name


def read_config(config_path):
    try:
        with open(config_path, 'r') as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

    except Exception as e:
        raise FileNotFoundError("There is no config at {}, yields {}".format(config_path, e))

    return config


def read_model(model_path, config):
    """
        Read keras retinanet model from keras.model.save()
        """
    #Suppress user warning, module does not need to be compiled for prediction
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        model = models.load_model(model_path, backbone_name='resnet50')

    return model


#Download progress bar
class DownloadProgressBar(tqdm):

    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def use_release(save_dir=os.path.join(_ROOT, "data/"), prebuilt_model="NEON"):
    '''Check the existance of, or download the latest model release from github
        
        Args:
                save_dir (str): Directory to save filepath, default to "data" in deepforest repo
                prebuilt_model: Currently only accepts "NEON", but could be expanded to include other prebuilt models. The local model will be called {prebuilt_model}.h5 on disk.
        
        Returns:
                release_tag, output_path (str): path to downloaded model 
        '''

    #Find latest github tag release from the DeepLidar repo
    _json = json.loads(
        urllib.request.urlopen(
            urllib.request.Request(
                'https://api.github.com/repos/Weecology/DeepForest/releases/latest',
                headers={'Accept': 'application/vnd.github.v3+json'},
            )).read())
    asset = _json['assets'][0]
    url = asset['browser_download_url']

    #Naming based on pre-built model
    output_path = os.path.join(save_dir, prebuilt_model + ".h5")

    #Check the release tagged locally
    try:
        release_txt = pd.read_csv(save_dir + "current_release.csv")
    except:
        release_txt = pd.DataFrame({"current_release": [None]})

    #Download the current release it doesn't exist
    if not release_txt.current_release[0] == _json["html_url"]:

        print("Downloading model from DeepForest release {}, see {} for details".format(
            _json["tag_name"], _json["html_url"]))

        with DownloadProgressBar(unit='B',
                                 unit_scale=True,
                                 miniters=1,
                                 desc=url.split('/')[-1]) as t:
            urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)

        print("Model was downloaded and saved to {}".format(output_path))

        #record the release tag locally
        release_txt = pd.DataFrame({"current_release": [_json["html_url"]]})
        release_txt.to_csv(save_dir + "current_release.csv")
    else:
        print(
            "Model from DeepForest release {} was already downloaded. Loading model from file."
            .format(_json["html_url"]))

    return _json["html_url"], output_path


def xml_to_annotations(xml_path):
    """Load annotations from xml format (e.g. RectLabel editor) and convert them into retinanet annotations format. 
        
        Args:
                xml_path (str): Path to the annotations xml, formatted by RectLabel
        
        Returns:
                Annotations (pandas dataframe): in the format -> path/to/image.png,x1,y1,x2,y2,class_name
        """
    #parse
    with open(xml_path) as fd:
        doc = xmltodict.parse(fd.read())

    #grab xml objects
    try:
        tile_xml = doc["annotation"]["object"]
    except Exception as e:
        raise Exception("error {} for path {} with doc annotation{}".format(
            e, xml_path, doc["annotation"]))

    xmin = []
    xmax = []
    ymin = []
    ymax = []
    label = []

    if type(tile_xml) == list:
        treeID = np.arange(len(tile_xml))

        #Construct frame if multiple trees
        for tree in tile_xml:
            xmin.append(tree["bndbox"]["xmin"])
            xmax.append(tree["bndbox"]["xmax"])
            ymin.append(tree["bndbox"]["ymin"])
            ymax.append(tree["bndbox"]["ymax"])
            label.append(tree['name'])
    else:
        #One tree
        treeID = 0

        xmin.append(tile_xml["bndbox"]["xmin"])
        xmax.append(tile_xml["bndbox"]["xmax"])
        ymin.append(tile_xml["bndbox"]["ymin"])
        ymax.append(tile_xml["bndbox"]["ymax"])
        label.append(tile_xml['name'])

    rgb_name = os.path.basename(doc["annotation"]["filename"])

    #set dtypes
    xmin = [int(x) for x in xmin]
    xmax = [int(x) for x in xmax]
    ymin = [int(x) for x in ymin]
    ymax = [int(x) for x in ymax]

    annotations = pd.DataFrame({
        "image_path": rgb_name,
        "xmin": xmin,
        "ymin": ymin,
        "xmax": xmax,
        "ymax": ymax,
        "label": label
    })
    return (annotations)


def create_classes(annotations_file):
    """Create a class list in the format accepted by keras retinanet
        
        Args:
                annotations_file: an annotation csv in the retinanet format path/to/image.png,x1,y1,x2,y2,class_name
        
        Returns:
                path to classes file
        """
    annotations = pd.read_csv(
        annotations_file, names=["image_path", "xmin", "ymin", "xmax", "ymax", "label"])

    #get dir to place along file annotations
    dirname = os.path.split(annotations_file)[0]
    classes_path = os.path.join(dirname, "classes.csv")

    #get unique labels
    labels = annotations.label.dropna().unique()
    n_classes = labels.shape[0]
    print("There are {} unique labels: {} ".format(n_classes, list(labels)))

    #write label
    with open(classes_path,'w', newline='\n', encoding='utf-8') as csv_file:
        writer = csv.writer(csv_file)
        for index, label in enumerate(labels):
            writer.writerow([label, index])

    return classes_path


def number_of_images(annotations_file):
    """How many images in the annotations file?
        
        Args:
                annotations_file (str):
        
        Returns:
                n (int): Number of images
        """

    df = pd.read_csv(annotations_file,
                     index_col=False,
                     names=["image_path", "xmin", "ymin", "xmax", "ymax"])
    n = len(df.image_path.unique())
    return n


def format_args(annotations_file, classes_file, config, images_per_epoch=None):
    """Format config file to match argparse list for retinainet
        
        Args:
                annotations_file: a path to a csv  dataframe of annotations to get number of images, no header
                config (dict): a dictionary object to convert into a list for argparse
                images_per_epoch (int): Override default steps per epoch (n images/batch size) by manually setting a number of images
        
        Returns:
                arg_list (list): a list structure that mimics argparse input arguments for retinanet
        """
    #Format args. Retinanet uses argparse, so they need to be passed as a list
    args = {}

    #remember that .yml reads None as a str
    if not config["weights"] == 'None':
        args["--weights"] = config["weights"]

    args["--backbone"] = config["backbone"]
    args["--image-min-side"] = config["image-min-side"]
    args["--multi-gpu"] = config["multi-gpu"]
    args["--epochs"] = config["epochs"]
    if images_per_epoch:
        args["--steps"] = round(images_per_epoch / int(config["batch_size"]))
    else:
        args["--steps"] = round(
            int(number_of_images(annotations_file)) / int(config["batch_size"]))

    args["--batch-size"] = config["batch_size"]
    args["--tensorboard-dir"] = None
    args["--workers"] = config["workers"]
    args["--max-queue-size"] = config["max_queue_size"]
    args["--freeze-layers"] = config["freeze_layers"]
    args["--score-threshold"] = config["score_threshold"]

    if config["save_path"]:
        args["--save-path"] = config["save_path"]

    if config["snapshot_path"]:
        args["--snapshot-path"] = config["snapshot_path"]

    arg_list = [[k, v] for k, v in args.items()]
    arg_list = [val for sublist in arg_list for val in sublist]

    #boolean arguments
    if config["save-snapshot"] is False:
        print("Disabling snapshot saving")
        arg_list = arg_list + ["--no-snapshots"]

    if config["freeze_resnet"] is True:
        arg_list = arg_list + ["--freeze-backbone"]

    if config["random_transform"] is True:
        print("Turning on random transform generator")
        arg_list = arg_list + ["--random-transform"]

    if config["multi-gpu"] > 1:
        arg_list = arg_list + ["--multi-gpu-force"]

    if config["multiprocessing"]:
        arg_list = arg_list + ["--multiprocessing"]

    #positional arguments first
    arg_list = arg_list + ["csv", annotations_file, classes_file]

    if not config["validation_annotations"] == "None":
        arg_list = arg_list + ["--val-annotations", config["validation_annotations"]]

    #All need to be str classes to mimic sys.arg
    arg_list = [str(x) for x in arg_list]

    return arg_list



#test module for tfrecords
from deepforest import tfrecords
from deepforest import utilities
from deepforest import preprocess
from deepforest import get_data

import pytest
import os
import glob
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt

from keras_retinanet.preprocessing import csv_generator
from keras_retinanet import models

@pytest.fixture()
def config():
    print("Configuring tfrecord tests")
    config = {}
    config["patch_size"] = 200
    config["patch_overlap"] = 0.05
    config["annotations_xml"] = get_data("OSBS_029.xml")
    config["rgb_dir"] = "data"
    config["annotations_file"] = "tests/data/OSBS_029.csv"
    config["path_to_raster"] =get_data("OSBS_029.tif")
    config["image-min-side"] = 800
    config["backbone"] = "resnet50"
    
    #Create a clean config test data
    annotations = utilities.xml_to_annotations(xml_path=config["annotations_xml"])
    annotations.to_csv("tests/data/tfrecords_OSBS_029.csv",index=False)
    
    annotations_file = preprocess.split_raster(path_to_raster=config["path_to_raster"],
                                                        annotations_file="tests/data/tfrecords_OSBS_029.csv",
                                                        base_dir= "tests/data/",
                                                        patch_size=config["patch_size"],
                                                        patch_overlap=config["patch_overlap"])
    
    annotations_file.to_csv("tests/data/testfile_tfrecords.csv", index=False,header=False)
    class_file = utilities.create_classes("tests/data/testfile_tfrecords.csv")    
    
    return config

#Reading

#Writing
@pytest.fixture()
def test_create_tfrecords(config):
    """This test is in flux due to the fact that tensorflow and cv2 resize methods are not identical: https://jricheimer.github.io/tensorflow/2019/02/11/resize-confusion/ """
    created_records = tfrecords.create_tfrecords(annotations_file="tests/data/testfile_tfrecords.csv",
                               class_file="tests/data/classes.csv",
                               image_min_side=config["image-min-side"], 
                               backbone_model=config["backbone"],
                               size=100,
                               savedir="tests/data/")
    assert os.path.exists("tests/data/testfile_tfrecords_0.tfrecord")
    return created_records

@pytest.fixture()
def setup_create_tensors(test_create_tfrecords):
    created_tensors = tfrecords.create_tensors(test_create_tfrecords)
    return created_tensors

def test_create_tensors(test_create_tfrecords):
    print("Testing that input tensors can be created")
    created_tensors = tfrecords.create_tensors(test_create_tfrecords)
    assert len(created_tensors) == 2
    
    return created_tensors

def test_create_dataset(test_create_tfrecords):
    dataset = tfrecords.create_dataset(test_create_tfrecords)
    
def test_equivalence(config, setup_create_tensors):
    
    #unpack created tensors
    tf_inputs, tf_targets = setup_create_tensors
    
    #the image going in to tensorflow should be equivalent to the image from the fit_generator
    backbone = models.backbone(config["backbone"])

    #CSV generator
    generator = csv_generator.CSVGenerator(
        csv_data_file="tests/data/testfile_tfrecords.csv",
        csv_class_file="tests/data/classes.csv",
        image_min_side=config["image-min-side"],
        preprocess_image=backbone.preprocess_image,
    )

    #find file in randomize generator group
    first_file = generator.groups[0][0]
    gen_filename = os.path.join(generator.base_dir, generator.image_names[first_file])
    original_image = generator.load_image(first_file)
    inputs, targets = generator.__getitem__(0)

    image = inputs[0,...]
    targets = targets[0][0,...]

    with tf.Session() as sess:
        #seek the randomized image to match
        tf_inputs, tf_targets = sess.run([tf_inputs,tf_targets])

    #assert filename is the same as generator
    #assert gen_filename == filename
    #tf_image = tf_image[0,...]
    tf_inputs = tf_inputs[0,...]
    tf_targets = tf_targets[0][0,...]

    #Same shape
    #assert tf_image.shape == image.shape
    assert tf_inputs.shape == image.shape
    assert tf_targets.shape == targets.shape

    #Same values, slightly duplicitious with above, but useful for debugging.
    #Saved array is the same as generator image
    #np.testing.assert_array_equal(image, tf_image)

    #Loaded array is the same as generator, this is not true currently, the opencv and the tensorflow interpolation method is slightly different, waiting for tf. 2.0
    #np.testing.assert_array_equal(tf_loaded, tf_image)

    ##Useful for debug to plot
    #fig = plt.figure()
    #ax1 = fig.add_subplot(1,4,1)
    #ax1.title.set_text('Fit Gen Original')    
    #plt.imshow(original_image[...,::-1])
    #ax1 = fig.add_subplot(1,4,2)
    #ax1.title.set_text('Fit Generator')    
    #plt.imshow(image)
    #ax2 = fig.add_subplot(1,4,3)
    #ax2 = fig.add_subplot(1,4,4)
    #ax2.title.set_text('Loaded Image')        
    #plt.imshow(tf_inputs)
    #plt.show()        

#Check for bad file types
#@pytest.fixture()
#def bad_annotations():
    #annotations = utilities.xml_to_annotations(get_data("OSBS_029.xml"))
    #f = "tests/data/testfile_error_deepforest.csv"
    #annotations.to_csv(f,index=False,header=False)
    #return f   

#def test_tfdataset_error(bad_annotations):    
    #with pytest.raises(ValueError):    
        #records_created = tfrecords.create_tfrecords(annotations_file=bad_annotations, class_file=get_data("classes.csv"), image_min_side=800, backbone_model="resnet50", size=100, savedir="tests/data/")


#test preprocessing
from PIL import Image
import numpy as np
import pytest
import pandas as pd
import os

from deepforest import preprocess
from deepforest import utilities
from deepforest import get_data

@pytest.fixture("module")
def config():
    config = utilities.read_config(get_data("deepforest_config.yml"))
    config["patch_size"] = 200
    config["patch_overlap"] = 0.25
    config["annotations_xml"] = get_data("OSBS_029.xml")
    config["rgb_dir"] = "data"
    config["annotations_file"] = "tests/data/OSBS_029.csv"
    config["path_to_raster"] = get_data("OSBS_029.tif")
    
    #Create a clean config test data
    annotations = utilities.xml_to_annotations(xml_path = config["annotations_xml"])
    annotations.to_csv("tests/data/OSBS_029.csv",index=False)
    
    return config

@pytest.fixture()
def numpy_image(config):
    raster = Image.open(config["path_to_raster"])
    return np.array(raster)
    
def test_compute_windows(config, numpy_image):
    windows = preprocess.compute_windows(numpy_image, config["patch_size"], config["patch_overlap"])
    assert len(windows) == 9

def test_select_annotations(config, numpy_image):      
    windows = preprocess.compute_windows(numpy_image, config["patch_size"], config["patch_overlap"])
    image_annotations = pd.read_csv("tests/data/OSBS_029.csv")
    selected_annotations = preprocess.select_annotations(image_annotations, windows, index=7)
    
    #Returns a 5 column matrix
    assert selected_annotations.shape[0] == 17
    
    #image name should be name of image plus the index .tif
    assert selected_annotations.image_path.unique()[0] == "OSBS_029_7.png"

def test_select_annotations_tile(config, numpy_image):
    config["patch_size"] = 50
    windows = preprocess.compute_windows(numpy_image, config["patch_size"], config["patch_overlap"])
    image_annotations = pd.read_csv("tests/data/OSBS_029.csv")    
    selected_annotations = preprocess.select_annotations(image_annotations, windows, index=10)
    
    #The largest box cannot be off the edge of the window
    assert selected_annotations.xmin.min() >= 0
    assert selected_annotations.ymin.min() >= 0
    assert selected_annotations.xmax.max() <= config["patch_size"]
    assert selected_annotations.ymax.max() <= config["patch_size"]
    
def test_split_raster(config):
    annotations_file = preprocess.split_raster(config["path_to_raster"], config["annotations_file"], "tests/data/",config["patch_size"], config["patch_overlap"])
    
    #Returns a 6 column pandas array
    assert annotations_file.shape[1] == 6
    
def test_split_raster_empty(config):
    #Blank annotations file
    blank_annotations = pd.DataFrame({"image_path":"OSBS_029.tif","xmin":[""],"ymin":[""],"xmax":[""],"ymax":[""],"label":[""]})
    blank_annotations.to_csv("tests/data/blank_annotations.csv",index=False)
    
    #Ignore blanks
    annotations_file = preprocess.split_raster(config["path_to_raster"], "tests/data/blank_annotations.csv", "tests/data/",config["patch_size"], config["patch_overlap"], allow_empty=False)
    assert annotations_file.shape[0] == 0
    
    #Include blanks
    annotations_file = preprocess.split_raster(config["path_to_raster"], "tests/data/blank_annotations.csv", "tests/data/",config["patch_size"], config["patch_overlap"], allow_empty=True)
    assert annotations_file.shape[0] > 0



#test data locations and existance
import os
import deepforest

#Make sure package data is present
def test_get_data():
    assert os.path.exists(deepforest.get_data("testfile_deepforest.csv"))
    assert os.path.exists(deepforest.get_data("OSBS_029.png"))
    assert os.path.exists(deepforest.get_data("OSBS_029.tif"))
    assert os.path.exists(deepforest.get_data("classes.csv"))


"""
Deepforest main module. This module holds the deepforest class for model building and training
"""
import os
import csv
import warnings

from PIL import Image
with warnings.catch_warnings():
    #Suppress some of the verbose tensorboard warnings, compromise to avoid numpy version errors
    warnings.filterwarnings("ignore", category=FutureWarning)
    import tensorflow as tf

import pandas as pd
import cv2
import numpy as np
from matplotlib import pyplot as plt

from deepforest import get_data
from deepforest import utilities
from deepforest import predict
from deepforest import preprocess
from deepforest.retinanet_train import main as retinanet_train
from deepforest.retinanet_train import parse_args

from keras_retinanet import models
from keras_retinanet.models import convert_model
from keras_retinanet.bin.train import create_models
from keras_retinanet.preprocessing.csv_generator import CSVGenerator, _read_classes
from keras_retinanet.utils.eval import evaluate
from keras_retinanet.utils.eval import _get_detections
from keras_retinanet.utils.visualization import draw_box


class deepforest:
    ''' 
    Class for training and predicting tree crowns in RGB images
    
    Args:
        weights (str): Path to model saved on disk from keras.model.save_weights(). A new model is created and weights are copied. Default is None. 
        saved_model: Path to a saved model from disk using keras.model.save(). No new model is created.
    
    Attributes:
        model: A keras training model from keras-retinanet
    '''

    def __init__(self, weights=None, saved_model=None):
        self.weights = weights
        self.saved_model = saved_model

        #Read config file - if a config file exists in local dir use it, if not use installed.
        if os.path.exists("deepforest_config.yml"):
            config_path = "deepforest_config.yml"
        else:
            try:
                config_path = get_data("deepforest_config.yml")
            except Exception as e:
                raise ValueError(
                    "No deepforest_config.yml found either in local directory or in installed package location. {}"
                    .format(e))

        print("Reading config file: {}".format(config_path))
        self.config = utilities.read_config(config_path)

        #Create a label dict, defaults to "Tree"
        self.read_classes()

        #release version id to flag if release is being used
        self.__release_version__ = None

        #Load saved model if needed
        if self.saved_model:
            print("Loading saved model")
            #Capture user warning, not relevant here
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning)
                self.model = utilities.load_model(saved_model)

        if self.weights is not None:
            print("Creating model from weights")
            backbone = models.backbone(self.config["backbone"])
            self.model, self.training_model, self.prediction_model = create_models(
                backbone.retinanet, num_classes=1, weights=self.weights)
        else:
            print(
                "A blank deepforest object created. To perform prediction, either train or load an existing model."
            )
            self.model = None

    def read_classes(self):
        """Read class file in case of multi-class training. If no file has been created, DeepForest assume there is 1 class, Tree"""
        # parse the provided class file
        self.labels = {}
        try:
            with open(self.classes_file, 'r') as file:
                self.classes = _read_classes(csv.reader(file, delimiter=','))
            for key, value in self.classes.items():
                self.labels[value] = key
        except:
            self.labels[0] = "Tree"

    def train(self,
              annotations,
              input_type="fit_generator",
              list_of_tfrecords=None,
              comet_experiment=None,
              images_per_epoch=None):
        '''Train a deep learning tree detection model using keras-retinanet.
        This is the main entry point for training a new model based on either existing weights or scratch
        
        Args:
            annotations (str): Path to csv label file, labels are in the format -> path/to/image.png,x1,y1,x2,y2,class_name
            comet_experiment: A comet ml object to log images. Optional.
            list_of_tfrecords: Ignored if input_type != "tfrecord", list of tf records to process
            input_type: "fit_generator" or "tfrecord"
            images_per_epoch: number of images to override default config of # images in annotations file / batch size. Useful for debug
        
        Returns:
            model (object): A trained keras model
            prediction model: with bbox nms
            trained model: without nms
        '''
        #Test if there is a new classes file in case # of classes has changed.
        self.classes_file = utilities.create_classes(annotations)
        self.read_classes()
        arg_list = utilities.format_args(annotations, self.classes_file, self.config,
                                         images_per_epoch)

        print("Training retinanet with the following args {}".format(arg_list))

        #Train model
        self.model, self.prediction_model, self.training_model = retinanet_train(
            forest_object=self,
            args=arg_list,
            input_type=input_type,
            list_of_tfrecords=list_of_tfrecords,
            comet_experiment=comet_experiment)

    def use_release(self, gpus=1):
        '''Use the latest DeepForest model release from github and load model. Optionally download if release doesn't exist
        
        Returns:
            model (object): A trained keras model
            gpus: number of gpus to parallelize, default to 1
        '''
        #Download latest model from github release
        release_tag, self.weights = utilities.use_release()

        #load saved model and tag release
        self.__release_version__ = release_tag
        print("Loading pre-built model: {}".format(release_tag))

        if gpus == 1:
            with warnings.catch_warnings():
                #Suppress compilte warning, not relevant here
                warnings.filterwarnings("ignore", category=UserWarning)
                self.model = utilities.read_model(self.weights, self.config)

            #Convert model
            self.prediction_model = convert_model(self.model)
        elif gpus > 1:
            backbone = models.backbone(self.config["backbone"])
            n_classes = len(self.labels.keys())
            self.model, self.training_model, self.prediction_model = create_models(
                backbone.retinanet,
                num_classes=n_classes,
                weights=self.weights,
                multi_gpu=gpus)

        #add to config
        self.config["weights"] = self.weights

    def predict_generator(self,
                          annotations,
                          comet_experiment=None,
                          iou_threshold=0.5,
                          max_detections=200,
                          return_plot=False):
        """Predict bounding boxes for a model using a csv fit_generator
    
        Args:
            annotations (str): Path to csv label file, labels are in the format -> path/to/image.png,x1,y1,x2,y2,class_name
            iou_threshold(float): IoU Threshold to count for a positive detection (defaults to 0.5)
            max_detections (int): Maximum number of bounding box predictions
            comet_experiment(object): A comet experiment class objects to track
            return_plot: Whether to return prediction boxes (False) or Images (True). If True, files will be written to current working directory if model.config["save_path"] is not defined.
        
        Return:
            boxes_output: If return_plot=False, a pandas dataframe of bounding boxes for each image in the annotations file
            None: If return_plot is True, images are written to save_dir as a side effect.
        """
        #Format args for CSV generator
        classes_file = utilities.create_classes(annotations)
        arg_list = utilities.format_args(annotations, classes_file, self.config)
        args = parse_args(arg_list)

        #create generator
        generator = CSVGenerator(
            args.annotations,
            args.classes,
            image_min_side=args.image_min_side,
            image_max_side=args.image_max_side,
            config=args.config,
            shuffle_groups=False,
        )

        if self.prediction_model:
            boxes_output = []

            #For each image, gather predictions
            for i in range(generator.size()):
                #pass image as path
                plot_name = generator.image_names[i]
                image_path = os.path.join(generator.base_dir, plot_name)
                result = self.predict_image(image_path,
                                            return_plot=return_plot,
                                            score_threshold=args.score_threshold)

                if return_plot:
                    if not self.config["save_path"]:
                        print(
                            "model.config['save_path'] is None, saving images to current working directory"
                        )
                        save_path = "."
                    else:
                        save_path = self.config["save_path"]
                    #Save image
                    fname = os.path.join(save_path, plot_name)
                    cv2.imwrite(fname, result)
                    continue
                else:
                    #Turn boxes to pandas frame and save output
                    box_df = pd.DataFrame(result)
                    #use only plot name, not extension
                    box_df["plot_name"] = os.path.splitext(plot_name)[0]
                    boxes_output.append(box_df)
        else:
            raise ValueError(
                "No prediction model loaded. Either load a retinanet from file, download the latest release or train a new model"
            )

        if return_plot:
            return None
        else:
            #if boxes, name columns and return box data
            boxes_output = pd.concat(boxes_output)
            boxes_output.columns = [
                "xmin", "ymin", "xmax", "ymax", "score", "label", "plot_name"
            ]
            boxes_output = boxes_output.reindex(
                columns=["plot_name", "xmin", "ymin", "xmax", "ymax", "score", "label"])
            return boxes_output

    def evaluate_generator(self,
                           annotations,
                           comet_experiment=None,
                           iou_threshold=0.5,
                           max_detections=200):
        """ Evaluate prediction model using a csv fit_generator
        
        Args:
            annotations (str): Path to csv label file, labels are in the format -> path/to/image.png,x1,y1,x2,y2,class_name
            iou_threshold(float): IoU Threshold to count for a positive detection (defaults to 0.5)
            max_detections (int): Maximum number of bounding box predictions
            comet_experiment(object): A comet experiment class objects to track
        
        Return:
            mAP: Mean average precision of the evaluated data
        """
        #Format args for CSV generator
        classes_file = utilities.create_classes(annotations)
        arg_list = utilities.format_args(annotations, classes_file, self.config)
        args = parse_args(arg_list)

        #create generator
        validation_generator = CSVGenerator(
            args.annotations,
            args.classes,
            image_min_side=args.image_min_side,
            image_max_side=args.image_max_side,
            config=args.config,
            shuffle_groups=False,
        )

        average_precisions = evaluate(validation_generator,
                                      self.prediction_model,
                                      iou_threshold=iou_threshold,
                                      score_threshold=args.score_threshold,
                                      max_detections=max_detections,
                                      save_path=args.save_path,
                                      comet_experiment=comet_experiment)

        # print evaluation
        total_instances = []
        precisions = []
        for label, (average_precision, num_annotations) in average_precisions.items():
            print('{:.0f} instances of class'.format(num_annotations),
                  validation_generator.label_to_name(label),
                  'with average precision: {:.4f}'.format(average_precision))
            total_instances.append(num_annotations)
            precisions.append(average_precision)

        if sum(total_instances) == 0:
            print('No test instances found.')
            return

        print('mAP using the weighted average of precisions among classes: {:.4f}'.format(
            sum([a * b for a, b in zip(total_instances, precisions)]) /
            sum(total_instances)))

        mAP = sum(precisions) / sum(x > 0 for x in total_instances)
        print('mAP: {:.4f}'.format(mAP))
        return mAP

    def predict_image(self,
                      image_path=None,
                      numpy_image=None,
                      return_plot=True,
                      score_threshold=0.05,
                      show=False,
                      color=None):
        """Predict tree crowns based on loaded (or trained) model
        
        Args:
            image_path (str): Path to image on disk
            numpy_image (array): Numpy image array in BGR channel order following openCV convention
            color (tuple): Color of bounding boxes in BGR order (0,0,0) black default 
            show (bool): Plot the predicted image with bounding boxes. Ignored if return_plot=False
            return_plot: Whether to return image with annotations overlaid, or just a numpy array of boxes
        
        Returns:
            predictions (array): if return_plot, an image. Otherwise a numpy array of predicted bounding boxes, with scores and labels
        """

        #Check for model save
        if (self.prediction_model is None):
            raise ValueError(
                "Model currently has no prediction weights, either train a new model using deepforest.train, loading existing model, or use prebuilt model (see deepforest.use_release()"
            )

        #Check the formatting
        if isinstance(image_path, np.ndarray):
            raise ValueError(
                "image_path should be a string, but is a numpy array. If predicting a loaded image (channel order BGR), use numpy_image argument."
            )

        #Check for correct formatting
        #Warning if image is very large and using the release model
        if numpy_image is None:
            numpy_image = cv2.imread(image_path)

        #Predict
        prediction = predict.predict_image(self.prediction_model,
                                           image_path=image_path,
                                           raw_image=numpy_image,
                                           return_plot=return_plot,
                                           score_threshold=score_threshold,
                                           color=color,
                                           classes=self.labels)

        #cv2 channel order to matplotlib order
        if return_plot & show:
            plt.imshow(prediction[:, :, ::-1])
            plt.show()

        return prediction

    def predict_tile(self,
                     raster_path=None,
                     numpy_image=None,
                     patch_size=400,
                     patch_overlap=0.15,
                     iou_threshold=0.15,
                     return_plot=False):
        """
        For images too large to input into the model, predict_tile cuts the image into overlapping windows, predicts trees on each window and reassambles into a single array. 
    
        Args:
            raster_path: Path to image on disk
            numpy_image (array): Numpy image array in BGR channel order following openCV convention
            iou_threshold: Minimum iou overlap among predictions between windows to be suppressed. Defaults to 0.5. Lower values suppress more boxes at edges.
            return_plot: Should the image be returned with the predictions drawn?
    
        Returns:
            boxes (array): if return_plot, an image. Otherwise a numpy array of predicted bounding boxes, scores and labels
        """

        if numpy_image:
            pass
        else:
            #Load raster as image
            raster = Image.open(raster_path)
            numpy_image = np.array(raster)

        #Compute sliding window index
        windows = preprocess.compute_windows(numpy_image, patch_size, patch_overlap)

        #Save images to tmpdir
        predicted_boxes = []

        for index, window in enumerate(windows):
            #Crop window and predict
            crop = numpy_image[windows[index].indices()]

            #Crop is RGB channel order, change to BGR
            crop = crop[..., ::-1]
            boxes = self.predict_image(numpy_image=crop,
                                       return_plot=False,
                                       score_threshold=self.config["score_threshold"])

            #transform coordinates to original system
            xmin, ymin, xmax, ymax = windows[index].getRect()
            boxes.xmin = boxes.xmin + xmin
            boxes.xmax = boxes.xmax + xmin
            boxes.ymin = boxes.ymin + ymin
            boxes.ymax = boxes.ymax + ymin

            predicted_boxes.append(boxes)

        predicted_boxes = pd.concat(predicted_boxes)

        #Non-max supression for overlapping boxes among window
        if patch_overlap == 0:
            mosaic_df = predicted_boxes
        else:
            with tf.Session() as sess:
                print("{} predictions in overlapping windows, applying non-max supression".
                      format(predicted_boxes.shape[0]))
                new_boxes, new_scores, new_labels = predict.non_max_suppression(
                    sess,
                    predicted_boxes[["xmin", "ymin", "xmax", "ymax"]].values,
                    predicted_boxes.score.values,
                    predicted_boxes.label.values,
                    max_output_size=predicted_boxes.shape[0],
                    iou_threshold=iou_threshold)
                
                #Recreate box dataframe
                image_detections = np.concatenate([
                    new_boxes,
                    np.expand_dims(new_scores, axis=1),
                    np.expand_dims(new_labels, axis=1)
                ],axis=1)
                
                mosaic_df = pd.DataFrame(
                    image_detections,
                    columns=["xmin", "ymin", "xmax", "ymax", "score", "label"])
                mosaic_df.label = mosaic_df.label.str.decode("utf-8")
                
                print("{} predictions kept after non-max suppression".format(
                    mosaic_df.shape[0]))

        if return_plot:
            #Draw predictions
            for box in mosaic_df[["xmin", "ymin", "xmax", "ymax"]].values:
                draw_box(numpy_image, box, [0, 0, 255])

            #Mantain consistancy with predict_image
            return numpy_image
        else:
            return mosaic_df

    def plot_curves(self):
        """Plot training curves"""
        if self.history:
            # Plot training & validation regression loss values
            fig, axes, = plt.subplots(nrows=1, ncols=3)
            axes = axes.flatten()

            #Regression Loss
            axes[0].plot(self.history.history['regression_loss'])
            axes[0].set_title('Bounding Box Loss')
            axes[0].set_ylabel('Loss')
            axes[0].set_xlabel('Epoch')

            #Classification Loss
            axes[1].plot(self.history.history['classification_loss'])
            axes[1].set_title('Classification Loss')
            axes[1].set_ylabel('Loss')
            axes[1].set_xlabel('Epoch')

            # Plot validation mAP
            if "mAP" in self.history.history.keys():
                axes[2].plot(self.history.history['mAP'])
                axes[2].set_title('Validation: Mean Average Precision')
                axes[2].set_ylabel('mAP')
                axes[2].set_xlabel('Epoch')
            plt.show()
        else:
            print("No training history found.")
            return None



"""
Tfrecord module
Tfrecords creation and reader for improved performance across multi-gpu
There were a tradeoffs made in this repo. It would be natural to save the generated prepreprocessed image to tfrecord from the generator. This results in enormous (100x) files. 
The compromise was to read the original image from file using tensorflow's data pipeline. The opencv resize billinear method is marginally different then the tensorflow method, so we can't literally assert they are the same array. 
"""
import tensorflow as tf
import os
import csv
import numpy as np

from math import ceil
import keras
import pandas as pd

from keras_retinanet.preprocessing.csv_generator import CSVGenerator
from keras_retinanet import models
from keras_retinanet.models.retinanet import retinanet_bbox
from keras_retinanet import losses

import matplotlib.pyplot as plt
import psutil
import gc


def create_tf_example(image, regression_target, class_target, fname, original_image):

    #Class labels can be stored as int list
    class_target = class_target.astype(int)

    #Save image information and metadata so that the tensors can be reshaped at runtime

    example = tf.train.Example(features=tf.train.Features(
        feature={
            'image/target_height':
                tf.train.Feature(int64_list=tf.train.Int64List(value=[image.shape[0]])),
            'image/target_width':
                tf.train.Feature(int64_list=tf.train.Int64List(value=[image.shape[1]])),
            'image/height':
                tf.train.Feature(int64_list=tf.train.Int64List(
                    value=[original_image.shape[0]])),
            'image/width':
                tf.train.Feature(int64_list=tf.train.Int64List(
                    value=[original_image.shape[1]])),
            'image/filename':
                tf.train.Feature(bytes_list=tf.train.BytesList(
                    value=[fname.encode('utf-8')])),
            'image/object/regression_target':
                tf.train.Feature(float_list=tf.train.FloatList(
                    value=regression_target.flatten())),
            'image/object/class_target':
                tf.train.Feature(int64_list=tf.train.Int64List(
                    value=class_target.flatten())),
            'image/object/n_anchors':
                tf.train.Feature(int64_list=tf.train.Int64List(
                    value=[regression_target.shape[0]]))
        }))

    # Serialize to string and write to file
    return example


def create_tfrecords(annotations_file,
                     class_file,
                     backbone_model="resnet50",
                     image_min_side=800,
                     size=1,
                     savedir="./"):
    """
    Args:
        annotations_file: path to 6 column data in form image_path, xmin, ymin, xmax, ymax, label
        backbone_model: A keras retinanet backbone
        image_min_side: resized image object minimum size
        size: Number of images per tfrecord
        savedir: dir path to save tfrecords files
    
    Returns:
        written_files: A list of path names of written tfrecords
    """
    memory_used = []

    #Image preprocess function
    backbone = models.backbone(backbone_model)

    #filebase name
    image_basename = os.path.splitext(os.path.basename(annotations_file))[0]

    ## Syntax checks
    ##Check annotations file only JPEG, PNG, GIF, or BMP are allowed.
    #df = pd.read_csv(annotations_file, names=["image_path","xmin","ymin","xmax","ymax","label"])
    #df['FileType'] = df.image_path.str.split('.').str[-1].str.lower()
    #bad_files = df[~df['FileType'].isin(["jpeg","jpg","png","gif","bmp"])]

    #if not bad_files.empty:
    #raise ValueError("Check annotations file, only JPEG, PNG, GIF, or BMP are allowed, {} incorrect files found /n {}: ".format(bad_files.shape[0],bad_files.head()))

    #Check dtypes, cannot use pandas, or will coerce in the presence of NAs
    with open(annotations_file, 'r') as f:
        reader = csv.reader(f, delimiter=',')
        row = next(reader)
        if row[1].count(".") > 0:
            raise ValueError(
                "Annotation files should be headerless with integer box, {} is not a int".
                format(row[1]))

    #Create generator - because of how retinanet yields data, this should always be 1. Shape problems in the future?
    train_generator = CSVGenerator(annotations_file,
                                   class_file,
                                   batch_size=1,
                                   image_min_side=image_min_side,
                                   preprocess_image=backbone.preprocess_image)

    #chunk size
    indices = np.arange(train_generator.size())
    chunks = [
        indices[i * size:(i * size) + size] for i in range(ceil(len(indices) / size))
    ]

    written_files = []
    for chunk in chunks:
        #Create tfrecord dataset and save it for output
        fname = savedir + "{}_{}.tfrecord".format(image_basename, chunk[0])
        written_files.append(fname)
        writer = tf.io.TFRecordWriter(fname)
        images = []
        regression_targets = []
        class_targets = []
        filename = []
        original_image = []
        for i in chunk:

            #Original image
            original_image.append(train_generator.load_image(i))

            batch = train_generator.__getitem__(i),

            #split into images and tar  gets
            inputs, targets = batch[0]

            #grab image, asssume batch size of 1, squeeze
            images.append(inputs[0, ...])

            #Grab anchor targets
            regression_batch, labels_batch = targets

            #grab regression anchors
            #regression_batch: batch that contains bounding-box regression targets for an image & anchor states (np.array of shape (batch_size, N, 4 + 1),
            #where N is the number of anchors for an image, the first 4 columns define regression targets for (x1, y1, x2, y2) and the
            #last column defines anchor states (-1 for ignore, 0 for bg, 1 for fg).
            regression_anchors = regression_batch[0, ...]
            regression_targets.append(regression_anchors)

            #grab class labels - squeeze out batch size
            #From retinanet: labels_batch: batch that contains labels & anchor states (np.array of shape (batch_size, N, num_classes + 1),
            #where N is the number of anchors for an image and the last column defines the anchor state (-1 for ignore, 0 for bg, 1 for fg).
            labels = labels_batch[0, ...]
            print("Label shape is: {}".format(labels.shape))
            class_targets.append(labels)

            #append filename by looking at group index
            current_index = train_generator.groups[i][0]

            #Grab filename and append to the full path
            fname = train_generator.image_names[current_index]
            fname = os.path.join(train_generator.base_dir, fname)

            filename.append(fname)

        for image, regression_target, class_target, fname, orig_image in zip(
                images, regression_targets, class_targets, filename, original_image):
            tf_example = create_tf_example(image, regression_target, class_target, fname,
                                           orig_image)
            writer.write(tf_example.SerializeToString())

        memory_used.append(psutil.virtual_memory().used / 2**30)

    plt.plot(memory_used)
    plt.title('Evolution of memory')
    plt.xlabel('iteration')
    plt.ylabel('memory used (GB)')
    plt.savefig(os.path.join(savedir, "memory.png"))

    return written_files


#Reading
def _parse_fn(example):

    #Define features
    features = {
        'image/filename': tf.io.FixedLenFeature([], tf.string),
        "image/object/regression_target": tf.VarLenFeature(tf.float32),
        "image/object/class_target": tf.VarLenFeature(tf.int64),
        "image/height": tf.FixedLenFeature([], tf.int64),
        "image/width": tf.FixedLenFeature([], tf.int64),
        "image/target_height": tf.FixedLenFeature([], tf.int64),
        "image/target_width": tf.FixedLenFeature([], tf.int64),
        "image/object/n_anchors": tf.FixedLenFeature([], tf.int64)
    }

    # Load one example and parse
    example = tf.io.parse_single_example(example, features)

    #Load image from file
    filename = tf.cast(example["image/filename"], tf.string)
    loaded_image = tf.read_file(filename)
    loaded_image = tf.image.decode_image(loaded_image, 3)

    #Reshape to known shape
    image_rows = tf.cast(example['image/height'], tf.int32)
    image_cols = tf.cast(example['image/width'], tf.int32)
    loaded_image = tf.reshape(loaded_image,
                              tf.stack([image_rows, image_cols, 3]),
                              name="cast_loaded_image")

    #needs to be float to subtract weights below
    loaded_image = tf.cast(loaded_image, tf.float32)

    #Turn loaded image from rgb into bgr and subtract imagenet means, see keras_retinanet.utils.image.preprocess_image
    red, green, blue = tf.unstack(loaded_image, axis=-1)

    #Subtract imagenet means
    blue = tf.subtract(blue, 103.939)
    green = tf.subtract(green, 116.779)
    red = tf.subtract(red, 123.68)

    #Recombine as BGR image
    loaded_image = tf.stack([blue, green, red], axis=-1)

    #Resize loaded image to desired target shape
    target_height = tf.cast(example['image/target_height'], tf.int32)
    target_width = tf.cast(example['image/target_width'], tf.int32)
    loaded_image = tf.image.resize(loaded_image, (target_height, target_width),
                                   align_corners=True)

    #Generated anchor data
    regression_target = tf.sparse_tensor_to_dense(
        example['image/object/regression_target'])
    class_target = tf.sparse_tensor_to_dense(example['image/object/class_target'])

    target_n_anchors = tf.cast(example['image/object/n_anchors'], tf.int32)
    regression_target = tf.cast(regression_target, tf.float32)
    class_target = tf.cast(class_target, tf.float32)

    regression_target = tf.reshape(regression_target, [target_n_anchors, 5],
                                   name="cast_regression")
    class_target = tf.reshape(class_target, [target_n_anchors, 2],
                              name="cast_class_label")

    return loaded_image, regression_target, class_target


def create_dataset(filepath, batch_size=1, shuffle=True):
    """
    Args:
        filepath: list of tfrecord files
        batch_size: number of images per batch
        
    Returns:
        dataset: a tensorflow dataset object for model training or prediction
    """

    # This works with arrays as well
    dataset = tf.data.TFRecordDataset(filepath)

    ## Set the number of datapoints you want to load and shuffle
    if shuffle:
        dataset = dataset.shuffle(800)

    ## This dataset will go on forever
    dataset = dataset.repeat()

    # Maps the parser on every filepath in the array. You can set the number of parallel loaders here
    dataset = dataset.map(_parse_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)

    ## Set the batchsize
    dataset = dataset.batch(batch_size=batch_size, drop_remainder=True)

    #Collect a queue of data tensors
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)

    ## Create an iterator
    iterator = dataset.make_one_shot_iterator()

    return iterator


def create_tensors(list_of_tfrecords, backbone_name="resnet50", shuffle=True):
    """
    Create a wired tensor target from a list of tfrecords
    
    Args:
        list_of_tfrecords: a list of tfrecord on disk to turn into a tfdataset
        backbone_name: keras retinanet backbone
        
    Returns:
        inputs: input tensors of images
        targets: target tensors of bounding boxes and classes
        """
    #Create tensorflow iterator
    iterator = create_dataset(list_of_tfrecords, shuffle=shuffle)
    next_element = iterator.get_next()

    #Split into inputs and targets
    inputs = next_element[0]
    targets = [next_element[1], next_element[2]]

    return inputs, targets



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `deepforest` package."""
import matplotlib
#matplotlib.use("MacOSX")

import os
import sys
import pytest
import keras
import glob
import numpy as np
from PIL import Image
from matplotlib import pyplot as plt

from deepforest import deepforest
from deepforest import preprocess
from deepforest import utilities
from deepforest import tfrecords
from deepforest import get_data

#download latest release
@pytest.fixture()
def download_release():
    print("running fixtures")
    utilities.use_release()    
    
@pytest.fixture()
def annotations():
    annotations = utilities.xml_to_annotations(get_data("OSBS_029.xml"))
    #Point at the png version for tfrecords
    annotations.image_path = annotations.image_path.str.replace(".tif",".png")
    
    annotations_file = get_data("testfile_deepforest.csv")
    annotations.to_csv(annotations_file,index=False,header=False)
    
    return annotations_file

@pytest.fixture()
def multi_annotations():
    annotations = utilities.xml_to_annotations(get_data("SOAP_061.xml"))
    annotations.image_path = annotations.image_path.str.replace(".tif",".png")
    annotations_file = get_data("testfile_multi.csv")
    annotations.to_csv(annotations_file,index=False,header=False)
    
    return annotations_file


@pytest.fixture()
def prepare_tfdataset(annotations):    
    records_created = tfrecords.create_tfrecords(annotations_file=annotations, class_file="tests/data/classes.csv", image_min_side=800, backbone_model="resnet50", size=100, savedir="tests/data/")
    assert os.path.exists("tests/data/testfile_deepforest_0.tfrecord")
    return records_created

def test_deepforest():
    model = deepforest.deepforest(weights=None)
    assert model.weights is None

def test_use_release(download_release):
    test_model = deepforest.deepforest() 
    test_model.use_release()
    
    #Check for release tag
    assert isinstance(test_model.__release_version__, str)
    
    #Assert is model instance    
    assert isinstance(test_model.model,keras.models.Model)
    
    assert test_model.config["weights"] == test_model.weights 
    assert test_model.config["weights"] is not "None"
    
@pytest.fixture()
def release_model(download_release):
    test_model = deepforest.deepforest() 
    test_model.use_release()
    
    #Check for release tag
    assert isinstance(test_model.__release_version__, str)
    
    #Assert is model instance    
    assert isinstance(test_model.model,keras.models.Model)
    
    return test_model

def test_predict_image(download_release):
    #Load model
    test_model = deepforest.deepforest(weights=get_data("NEON.h5"))    
    assert isinstance(test_model.model,keras.models.Model)
    
    #Predict test image and return boxes
    boxes = test_model.predict_image(image_path=get_data("OSBS_029.tif"), show=False, return_plot = False, score_threshold=0.1)
    
    #Returns a 6 column numpy array, xmin, ymin, xmax, ymax, score, label
    assert boxes.shape[1] == 6
    
    assert boxes.score.min() > 0.1

@pytest.fixture()
def test_train(annotations):
    test_model = deepforest.deepforest()
    test_model.config["epochs"] = 1
    test_model.config["save-snapshot"] = False
    test_model.config["steps"] = 1
    test_model.train(annotations=annotations, input_type="fit_generator")
    
    return test_model

def test_predict_generator(release_model, annotations):
    release_model.config["save_path"] = "tests/output/"
    boxes = release_model.predict_generator(annotations=annotations, return_plot=False)
    assert boxes.shape[1] == 7
    
    release_model.predict_generator(annotations=annotations, return_plot=True)    
    assert os.path.exists("tests/output/OSBS_029.png")
   
def test_evaluate_generator(release_model, annotations):
    mAP = release_model.evaluate_generator(annotations=annotations)
    assert isinstance(mAP, float)
    
#Test random transform
def test_random_transform(annotations):
    test_model = deepforest.deepforest()
    test_model.config["random_transform"] = True
    classes_file = utilities.create_classes(annotations)
    arg_list = utilities.format_args(annotations, classes_file, test_model.config)
    assert "--random-transform" in arg_list

def test_predict_tile(release_model):
    raster_path = get_data("OSBS_029.tif")
    image = release_model.predict_tile(raster_path,patch_size=300,patch_overlap=0.5,return_plot=True)

    #Test no non-max suppression
    boxes = release_model.predict_tile(raster_path,patch_size=100,patch_overlap=0,return_plot=False)
    assert not boxes.empty
    
def test_retrain_release(annotations, release_model):
    release_model.config["epochs"] = 1
    release_model.config["save-snapshot"] = False
    release_model.config["steps"] = 1
    
    assert release_model.config["weights"] == release_model.weights
    
    #test that it gets passed to retinanet
    classes_file = utilities.create_classes(annotations)
    arg_list = utilities.format_args(annotations, classes_file, release_model.config, images_per_epoch=1)
    strs = ["--weights" == x for x in arg_list]
    index = np.where(strs)[0][0] + 1
    assert arg_list[index] == release_model.weights
    
    release_model.train(annotations=annotations, input_type="fit_generator")
    

def test_multi_train(multi_annotations):    
    test_model = deepforest.deepforest()
    test_model.config["epochs"] = 1
    test_model.config["save-snapshot"] = False
    test_model.config["steps"] = 1
    test_model.train(annotations=multi_annotations, input_type="fit_generator")
    
    # Test labels
    labels = list(test_model.labels.values())
    labels.sort()
    target_labels = ["Dead","Alive"]
    target_labels.sort()
    
    assert labels == target_labels


#Deepforest Preprocessing model
'''
The preprocessing module is used to reshape data into format suitable for training or prediction. For example cutting large tiles into smaller images.
'''
import pandas as pd
from PIL import Image
import numpy as np
import os
import slidingwindow


def image_name_from_path(image_path):
    """
    Convert path to image name for use in indexing
    """
    image_name = os.path.basename(image_path)
    image_name = os.path.splitext(image_name)[0]

    return image_name


def compute_windows(numpy_image, patch_size, patch_overlap):
    ''''Create a sliding window object from a raster tile
    
    Args:
        numpy_image (array): Raster object as numpy array to cut into crops
    
    Returns:
        windows (list): a sliding windows object
    '''

    if patch_overlap > 1:
        raise ValueError("Patch overlap {} must be between 0 - 1".format(patch_overlap))

    #Generate overlapping sliding windows
    windows = slidingwindow.generate(numpy_image,
                                     slidingwindow.DimOrder.HeightWidthChannel,
                                     patch_size, patch_overlap)

    return (windows)


def select_annotations(annotations, windows, index, allow_empty=False):
    """Select annotations that overlap with selected image crop
    
    Args:
        image_name (str): Name of the image in the annotations file to lookup. 
        annotations_file: path to annotations file in the format -> image_path, xmin, ymin, xmax, ymax, label
        windows: A sliding window object (see compute_windows) 
        index: The index in the windows object to use a crop bounds
        allow_empty (bool): If True, allow window crops that have no annotations to be included
    
    Returns:
        selected_annotations: a pandas dataframe of annotations 
    """

    # Window coordinates - with respect to tile
    window_xmin, window_ymin, w, h = windows[index].getRect()
    window_xmax = window_xmin + w
    window_ymax = window_ymin + h

    #buffer coordinates a bit to grab boxes that might start just against the image edge. Don't allow boxes that start and end after the offset
    offset = 40
    selected_annotations = annotations[(annotations.xmin > (window_xmin - offset)) &
                                       (annotations.xmin < (window_xmax)) &
                                       (annotations.xmax >
                                        (window_xmin)) & (annotations.ymin >
                                                          (window_ymin - offset)) &
                                       (annotations.xmax <
                                        (window_xmax + offset)) & (annotations.ymin <
                                                                   (window_ymax)) &
                                       (annotations.ymax >
                                        (window_ymin)) & (annotations.ymax <
                                                          (window_ymax + offset))].copy()

    #change the image name
    image_name = os.path.splitext("{}".format(annotations.image_path.unique()[0]))[0]
    image_basename = os.path.splitext(image_name)[0]
    selected_annotations.image_path = "{}_{}.png".format(image_basename, index)

    ##If no matching annotations, return a line with the image name, but no records
    if selected_annotations.empty:
        if allow_empty:
            selected_annotations = pd.DataFrame(
                ["{}_{}.png".format(image_basename, index)], columns=["image_path"])
            selected_annotations["xmin"] = ""
            selected_annotations["ymin"] = ""
            selected_annotations["xmax"] = ""
            selected_annotations["ymax"] = ""
            selected_annotations["label"] = ""
        else:
            pass
    else:
        #update coordinates with respect to origin
        selected_annotations.xmax = (selected_annotations.xmin - window_xmin) + (
            selected_annotations.xmax - selected_annotations.xmin)
        selected_annotations.xmin = (selected_annotations.xmin - window_xmin)
        selected_annotations.ymax = (selected_annotations.ymin - window_ymin) + (
            selected_annotations.ymax - selected_annotations.ymin)
        selected_annotations.ymin = (selected_annotations.ymin - window_ymin)

        #cut off any annotations over the border.
        selected_annotations.loc[selected_annotations.xmin < 0, "xmin"] = 0
        selected_annotations.loc[selected_annotations.xmax > w, "xmax"] = w
        selected_annotations.loc[selected_annotations.ymin < 0, "ymin"] = 0
        selected_annotations.loc[selected_annotations.ymax > h, "ymax"] = h

    return selected_annotations


def save_crop(base_dir, image_name, index, crop):
    """
    Save window crop as image file to be read by PIL. Filename should match the image_name + window index

    """
    #create dir if needed
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    im = Image.fromarray(crop)
    image_basename = os.path.splitext(image_name)[0]
    filename = "{}/{}_{}.png".format(base_dir, image_basename, index)
    im.save(filename)

    return filename


def split_raster(path_to_raster,
                 annotations_file,
                 base_dir=".",
                 patch_size=400,
                 patch_overlap=0.05,
                 allow_empty=False):
    """Divide a large tile into smaller arrays. Each crop will be saved to file
    
    Args:
        path_to_tile (str): Path to a tile that can be read by rasterio on disk
        annotations_file (str): Path to annotations file (with column names) data in the format -> image_path, xmin, ymin, xmax, ymax, label
        base_dir (str): Where to save the annotations and image crops relative to current working dir
        patch_size (int): Maximum dimensions of square window
        patch_overlap (float): Percent of overlap among windows 0->1
        allow_empty: If True, include images with no annotations to be included in the dataset
        
    Returns:
        A pandas dataframe with annotations file for training.
    """
    #Load raster as image
    raster = Image.open(path_to_raster)
    numpy_image = np.array(raster)

    #Check that its 3 band
    bands = numpy_image.shape[2]
    if not bands == 3:
        raise IOError(
            "Input file {} has {} bands. DeepForest only accepts 3 band RGB rasters. If the image was cropped and saved as a .jpg, please ensure that no alpha channel was used."
            .format(path_to_raster, bands))

    #Compute sliding window index
    windows = compute_windows(numpy_image, patch_size, patch_overlap)

    #Get image name for indexing
    image_name = os.path.basename(path_to_raster)

    # Load annotations file and coerce dtype
    annotations = pd.read_csv(annotations_file)

    #open annotations file
    image_annotations = annotations[annotations.image_path == image_name].copy()

    #Sanity checks
    if image_annotations.empty:
        raise ValueError(
            "No image names match between the file:{} and the image_path: {}. Reminder that image paths should be the relative path (e.g. 'image_name.tif'), not the full path (e.g. path/to/dir/image_name.tif)"
            .format(annotations_file, image_name))

    if not annotations.shape[1] == 6:
        raise ValueError(
            "Annotations file has {} columns, should have format image_path, xmin, ymin, xmax, ymax, label"
            .format(annotations.shape[1]))

    annotations_files = []
    for index, window in enumerate(windows):

        #Crop image
        crop = numpy_image[windows[index].indices()]

        #Find annotations, image_name is the basename of the path
        crop_annotations = select_annotations(image_annotations, windows, index,
                                              allow_empty)

        #save annotations
        annotations_files.append(crop_annotations)

        #save image crop
        save_crop(base_dir, image_name, index, crop)

    annotations_files = pd.concat(annotations_files)

    #Checkpoint csv files, useful for parallelization
    #Use filename of the raster path to save the annotations
    image_basename = os.path.splitext(image_name)[0]
    file_path = image_basename + ".csv"
    file_path = os.path.join(base_dir, file_path)
    annotations_files.to_csv(file_path, index=False, header=False)

    return annotations_files



#test loading of keras retinanet
import os

def test_keras_retinanet():
    import keras_retinanet
    
def test_Cython_build():
    import keras_retinanet.utils.compute_overlap  
    assert os.path.exists(keras_retinanet.utils.compute_overlap.__file__)


"""
Prediction module. This module consists of predict utility function for the deepforest class
"""
import numpy as np
import copy
import glob
import keras
import cv2
import pandas as pd
import tensorflow as tf

#Retinanet-viz
from keras_retinanet.utils import image as keras_retinanet_image
from keras_retinanet.utils.visualization import draw_detections

def predict_image(model,
                  image_path=None,
                  raw_image=None,
                  score_threshold=0.05,
                  max_detections=200,
                  return_plot=True,
                  classes={"0": "Tree"},
                  color=None):
    """
    Predict invidiual tree crown bounding boxes for a single image
    
    Args:
        model (object): A keras-retinanet model to predict bounding boxes, either load a model from weights, use the latest release, or train a new model from scratch.  
        image_path (str): Path to image file on disk
        raw_image (str): Numpy image array in BGR channel order following openCV convention
        score_threshold (float): Minimum probability score to be included in final boxes, ranging from 0 to 1.
        max_detections (int): Maximum number of bounding box predictions per tile
        return_plot (bool):  If true, return a image object, else return bounding boxes as a numpy array
    
    Returns:
        raw_image (array): If return_plot is TRUE, the image with the overlaid boxes is returned
        image_detections: If return_plot is FALSE, a np.array of image_boxes, image_scores, image_labels
    """
    #Check for raw_image
    if raw_image is not None:
        numpy_image = raw_image.copy()
    else:
        #Read from path
        numpy_image = cv2.imread(image_path)

    #Make sure image exists
    try:
        numpy_image.shape
    except:
        raise IOError(
            "Image file {} cannot be read, check that it exists".format(image_path))

    #Check that its 3 band
    bands = numpy_image.shape[2]
    if not bands == 3:
        raise IOError(
            "Input file {} has {} bands. DeepForest only accepts 3 band RGB rasters. If the image was cropped and saved as a .jpg, please ensure that no alpha channel was used."
            .format(image_path, bands))

    image = keras_retinanet_image.preprocess_image(numpy_image)
    image, scale = keras_retinanet_image.resize_image(image)

    if keras.backend.image_data_format() == 'channels_first':
        image = image.transpose((2, 0, 1))

    # run network
    boxes, scores, labels = model.predict_on_batch(np.expand_dims(image, axis=0))[:3]

    # correct boxes for image scale
    boxes /= scale

    # select indices which have a score above the threshold
    indices = np.where(scores[0, :] > score_threshold)[0]

    # select those scores
    scores = scores[0][indices]

    # find the order with which to sort the scores
    scores_sort = np.argsort(-scores)[:max_detections]

    # select detections
    image_boxes = boxes[0, indices[scores_sort], :]
    image_scores = scores[scores_sort]
    image_labels = labels[0, indices[scores_sort]]
    image_detections = np.concatenate([
        image_boxes,
        np.expand_dims(image_scores, axis=1),
        np.expand_dims(image_labels, axis=1)
    ],
                                      axis=1)

    df = pd.DataFrame(image_detections,
                      columns=["xmin", "ymin", "xmax", "ymax", "score", "label"])

    #Change numberic class into string label
    df.label = df.label.astype(int)
    df.label = df.label.apply(lambda x: classes[x])

    if return_plot:
        draw_detections(numpy_image,
                        image_boxes,
                        image_scores,
                        image_labels,
                        label_to_name=None,
                        score_threshold=score_threshold,
                        color=color)
        return numpy_image
    else:
        return df


def non_max_suppression(sess,
                        boxes,
                        scores,
                        labels,
                        max_output_size=200,
                        iou_threshold=0.15):
    '''
    Provide a tensorflow session and get non-maximum suppression
    Args:
        sess: a tensorfloe
    max_output_size, iou_threshold are passed to tf.image.non_max_suppression 
    '''
    non_max_idxs = tf.image.non_max_suppression(boxes,
                                                scores,
                                                max_output_size=max_output_size,
                                                iou_threshold=iou_threshold)
    new_boxes = tf.cast(tf.gather(boxes, non_max_idxs), tf.int32)
    new_scores = tf.gather(scores, non_max_idxs)
    new_labels = tf.gather(labels, non_max_idxs)
    return sess.run([new_boxes, new_scores, new_labels])



#Retinanet training
"""
Retinanet training module.
Developed from keras-retinanet repo
"""

import argparse
import os
import sys
import warnings

import keras
import keras.preprocessing.image
import tensorflow as tf

#Retinanet
from keras_retinanet import layers
from keras_retinanet import losses
from keras_retinanet import models
from keras_retinanet.callbacks import RedirectModel
from keras_retinanet.callbacks.eval import Evaluate
from keras_retinanet.models.retinanet import retinanet_bbox
from keras_retinanet.preprocessing.csv_generator import CSVGenerator
from keras_retinanet.utils.anchors import make_shapes_callback
from keras_retinanet.utils.config import read_config_file, parse_anchor_parameters
from keras_retinanet.utils.keras_version import check_keras_version
from keras_retinanet.utils.model import freeze as freeze_model
from keras_retinanet.utils.transform import random_transform_generator
from keras_retinanet.utils.image import random_visual_effect_generator
from keras_retinanet.utils.gpu import setup_gpu

from deepforest import tfrecords


def makedirs(path):
    # Intended behavior: try to create the directory,
    # pass if the directory exists already, fails otherwise.
    # Meant for Python 2.7/3.n compatibility.
    try:
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise


def model_with_weights(model, weights, skip_mismatch):
    """ Load weights for model.

    Args:
        model         : The model to load weights for.
        weights       : The weights to load.
        skip_mismatch : If True, skips layers whose shape of weights doesn't match with the model.
    """
    if weights is not None:
        model.load_weights(weights, by_name=True, skip_mismatch=skip_mismatch)
    return model


def create_models(backbone_retinanet,
                  num_classes,
                  weights,
                  multi_gpu=0,
                  freeze_backbone=False,
                  lr=1e-5,
                  config=None,
                  targets=None,
                  freeze_layers=0,
                  modifier=None):
    """ Creates three models (model, training_model, prediction_model).

    Args:
        backbone_retinanet : A function to call to create a retinanet model with a given backbone.
        num_classes        : The number of classes to train.
        weights            : The weights to load into the model.
        multi_gpu          : The number of GPUs to use for training.
        freeze_backbone    : If True, disables learning for the backbone.
        config             : Config parameters, None indicates the default configuration.
        targets            : Target tensors if training a model with tfrecord inputs
        freeze_layers    : int layer number to freeze from bottom of the retinanet network during finetuning. e.g. 10 will set layers 0:10 to layer.trainable = False. 0 is default, no freezing.
        modifier           : function that takes in a model and freezes resnet layers, returns modified object

    Returns:
        model            : The base model. This is also the model that is saved in snapshots.
        training_model   : The training model. If multi_gpu=0, this is identical to model.
        prediction_model : The model wrapped with utility functions to perform object detection (applies regression values and performs NMS).
    """

    #if not modifier:
    modifier = freeze_model if freeze_backbone else None

    # load anchor parameters, or pass None (so that defaults will be used)
    anchor_params = None
    num_anchors = None
    if config and 'anchor_parameters' in config:
        anchor_params = parse_anchor_parameters(config)
        num_anchors = anchor_params.num_anchors()

    # Keras recommends initialising a multi-gpu model on the CPU to ease weight sharing, and to prevent OOM errors.
    # optionally wrap in a parallel model
    if multi_gpu > 1:
        from keras.utils import multi_gpu_model
        with tf.device('/cpu:0'):
            model = model_with_weights(backbone_retinanet(num_classes,
                                                          num_anchors=num_anchors,
                                                          modifier=modifier),
                                       weights=weights,
                                       skip_mismatch=True)
        training_model = multi_gpu_model(model, gpus=multi_gpu)
    else:
        model = model_with_weights(backbone_retinanet(num_classes,
                                                      num_anchors=num_anchors,
                                                      modifier=modifier),
                                   weights=weights,
                                   skip_mismatch=True)
        training_model = model

    # make prediction model
    prediction_model = retinanet_bbox(model=model, anchor_params=anchor_params)

    #Compile model
    if targets:
        #tfdataset target tensor from tfrecords pipelione
        training_model.compile(loss={
            'regression': losses.smooth_l1(),
            'classification': losses.focal()
        },
                               optimizer=keras.optimizers.adam(lr=lr, clipnorm=0.001),
                               target_tensors=targets)
    else:
        # compile model
        training_model.compile(loss={
            'regression': losses.smooth_l1(),
            'classification': losses.focal()
        },
                               optimizer=keras.optimizers.adam(lr=lr, clipnorm=0.001))

    return model, training_model, prediction_model


def create_callbacks(model, training_model, prediction_model, validation_generator, args,
                     comet_experiment):
    """ Creates the callbacks to use during training.

    Args
        model: The base model.
        training_model: The model that is used for training.
        prediction_model: The model that should be used for validation.
        validation_generator: The generator for creating validation data.
        args: parseargs args object.
        comet_experiment: cometml object to log images

    Returns:
        A list of callbacks used for training.
    """
    callbacks = []

    if args.evaluation and validation_generator:
        evaluation = Evaluate(validation_generator,
                              weighted_average=args.weighted_average,
                              comet_experiment=comet_experiment,
                              save_path=args.save_path,
                              score_threshold=args.score_threshold)
        evaluation = RedirectModel(evaluation, prediction_model)
        callbacks.append(evaluation)

    # save the model
    if args.snapshots:
        # ensure directory created first; otherwise h5py will error after epoch.
        makedirs(args.snapshot_path)
        checkpoint = keras.callbacks.ModelCheckpoint(
            os.path.join(
                args.snapshot_path, '{backbone}_{dataset_type}_{{epoch:02d}}.h5'.format(
                    backbone=args.backbone, dataset_type=args.dataset_type)),
            verbose=1,
            # save_best_only=True,
            # monitor="mAP",
            # mode='max'
        )
        checkpoint = RedirectModel(checkpoint, model)
        callbacks.append(checkpoint)

    callbacks.append(
        keras.callbacks.ReduceLROnPlateau(monitor='loss',
                                          factor=0.1,
                                          patience=2,
                                          verbose=1,
                                          mode='auto',
                                          min_delta=0.0001,
                                          cooldown=0,
                                          min_lr=0))

    return callbacks


def create_generators(args, preprocess_image):
    """ Create generators for training and validation.

    Args:
        args             : parseargs object containing configuration for generators.
        preprocess_image : Function that preprocesses an image for the network.
    """
    common_args = {
        'batch_size': args.batch_size,
        'config': args.config,
        'image_min_side': args.image_min_side,
        'image_max_side': args.image_max_side,
        'preprocess_image': preprocess_image,
    }

    # create random transform generator for augmenting training data
    if args.random_transform:
        transform_generator = random_transform_generator(
            min_rotation=-0.1,
            max_rotation=0.1,
            min_translation=(-0.1, -0.1),
            max_translation=(0.1, 0.1),
            min_shear=-0.1,
            max_shear=0.1,
            min_scaling=(0.9, 0.9),
            max_scaling=(1.1, 1.1),
            flip_x_chance=0.5,
            flip_y_chance=0.5,
        )
        visual_effect_generator = random_visual_effect_generator(
            contrast_range=(0.9, 1.1),
            brightness_range=(-.1, .1),
            hue_range=(-0.05, 0.05),
            saturation_range=(0.95, 1.05))
    else:
        transform_generator = random_transform_generator(flip_x_chance=0.5)
        visual_effect_generator = None
    if args.dataset_type == 'csv':
        train_generator = CSVGenerator(args.annotations,
                                       args.classes,
                                       transform_generator=transform_generator,
                                       visual_effect_generator=visual_effect_generator,
                                       **common_args)

        if args.val_annotations:
            validation_generator = CSVGenerator(args.val_annotations,
                                                args.classes,
                                                shuffle_groups=False,
                                                **common_args)
        else:
            validation_generator = None
    else:
        raise ValueError('Invalid data type received: {}'.format(args.dataset_type))

    return train_generator, validation_generator


def check_args(parsed_args):
    """ Function to check for inherent contradictions within parsed arguments.
    For example, batch_size < num_gpus
    Intended to raise errors prior to backend initialisation.

    Args:
        parsed_args: parser.parse_args()

    Returns:
        parsed_args
    """

    if parsed_args.multi_gpu > 1 and parsed_args.batch_size < parsed_args.multi_gpu:
        raise ValueError(
            "Batch size ({}) must be equal to or higher than the number of GPUs ({})".
            format(parsed_args.batch_size, parsed_args.multi_gpu))

    if parsed_args.multi_gpu > 1 and parsed_args.snapshot:
        raise ValueError(
            "Multi GPU training ({}) and resuming from snapshots ({}) is not supported.".
            format(parsed_args.multi_gpu, parsed_args.snapshot))

    if parsed_args.multi_gpu > 1 and not parsed_args.multi_gpu_force:
        raise ValueError(
            "Multi-GPU support is experimental, use at own risk! Run with --multi-gpu-force if you wish to continue."
        )

    if 'resnet' not in parsed_args.backbone:
        warnings.warn(
            'Using experimental backbone {}. Only resnet50 has been properly tested.'.
            format(parsed_args.backbone))

    return parsed_args


def parse_args(args):
    """ Parse the arguments.
    """
    parser = argparse.ArgumentParser(
        description='Simple training script for training a RetinaNet network.')
    subparsers = parser.add_subparsers(help='Arguments for specific dataset types.',
                                       dest='dataset_type')
    subparsers.required = True

    def csv_list(string):
        return string.split(',')

    csv_parser = subparsers.add_parser('csv')
    csv_parser.add_argument('annotations',
                            help='Path to CSV file containing annotations for training.')
    csv_parser.add_argument('classes',
                            help='Path to a CSV file containing class label mapping.')
    csv_parser.add_argument(
        '--val-annotations',
        help='Path to CSV file containing annotations for validation (optional).')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--snapshot', help='Resume training from a snapshot.')
    group.add_argument(
        '--imagenet-weights',
        help=
        'Initialize the model with pretrained imagenet weights. This is the default behaviour.',
        action='store_const',
        const=True,
        default=True)
    group.add_argument('--weights', help='Initialize the model with weights from a file.')
    group.add_argument('--no-weights',
                       help='Don\'t initialize the model with any weights.',
                       dest='imagenet_weights',
                       action='store_const',
                       const=False)

    parser.add_argument('--backbone',
                        help='Backbone model used by retinanet.',
                        default='resnet50',
                        type=str)
    parser.add_argument('--batch-size', help='Size of the batches.', default=1, type=int)
    parser.add_argument('--gpu', help='Id of the GPU to use (as reported by nvidia-smi).')
    parser.add_argument('--multi-gpu',
                        help='Number of GPUs to use for parallel processing.',
                        type=int,
                        default=0)
    parser.add_argument(
        '--multi-gpu-force',
        help='Extra flag needed to enable (experimental) multi-gpu support.',
        action='store_true')
    parser.add_argument('--epochs',
                        help='Number of epochs to train.',
                        type=int,
                        default=50)
    parser.add_argument('--steps',
                        help='Number of steps per epoch.',
                        type=int,
                        default=10000)
    parser.add_argument('--lr', help='Learning rate.', type=float, default=1e-5)
    parser.add_argument(
        '--snapshot-path',
        help=
        'Path to store snapshots of models during training (defaults to \'./snapshots\')',
        default='./snapshots')
    parser.add_argument(
        '--save-path',
        help='Path to store images during callback (defaults to \'./snapshots\')',
        default='./snapshots')
    parser.add_argument('--tensorboard-dir',
                        help='Log directory for Tensorboard output',
                        default='./logs')
    parser.add_argument('--no-snapshots',
                        help='Disable saving snapshots.',
                        dest='snapshots',
                        action='store_false')
    parser.add_argument('--no-evaluation',
                        help='Disable per epoch evaluation.',
                        dest='evaluation',
                        action='store_false')
    parser.add_argument('--freeze-backbone',
                        help='Freeze training of backbone layers.',
                        action='store_true')
    parser.add_argument('--random-transform',
                        help='Randomly transform image and annotations.',
                        action='store_true')
    parser.add_argument('--image-min-side',
                        help='Rescale the image so the smallest side is min_side.',
                        type=int,
                        default=1000)
    parser.add_argument(
        '--image-max-side',
        help='Rescale the image if the largest side is larger than max_side.',
        type=int,
        default=1333)
    parser.add_argument('--config', help='Path to a configuration parameters .ini file.')
    parser.add_argument(
        '--weighted-average',
        help='Compute the mAP using the weighted average of precisions among classes.',
        action='store_true')
    parser.add_argument('--compute-val-loss',
                        help='Compute validation loss during training',
                        dest='compute_val_loss',
                        action='store_true')
    parser.add_argument('--freeze-layers',
                        help='Set trainable to False for the first n retinanet layers',
                        type=int,
                        default=0)

    # Fit generator arguments
    parser.add_argument('--multiprocessing',
                        help='Use multiprocessing in fit_generator.',
                        action='store_true')
    parser.add_argument('--workers',
                        help='Number of generator workers.',
                        type=int,
                        default=1)
    parser.add_argument('--max-queue-size',
                        help='Queue length for multiprocessing workers in fit_generator.',
                        type=int,
                        default=10)

    #callback arguments
    parser.add_argument('--score-threshold',
                        help="Minimum bounding box score to be considered in prediction",
                        type=float,
                        default=0.05)

    return check_args(parser.parse_args(args))


def main(forest_object,
         args=None,
         input_type="fit_generator",
         list_of_tfrecords=None,
         comet_experiment=None):
    """
    Main Training Loop
    Args:
        forest_object: a deepforest class object
        args: Keras retinanet argparse
        list_of_tfrecords: list of tfrecords to parse
        input_type: "fit_generator" or "tfrecord" input type
    """
    # parse arguments
    if args is None:
        args = sys.argv[1:]
    args = parse_args(args)

    # create object that stores backbone information
    backbone = models.backbone(args.backbone)

    # make sure keras is the minimum required version
    check_keras_version()

    # optionally choose specific GPU
    if args.gpu:
        setup_gpu(args.gpu)

    # optionally load config parameters
    if args.config:
        args.config = read_config_file(args.config)

    #data input
    if input_type == "fit_generator":
        # create the generators
        train_generator, validation_generator = create_generators(
            args, backbone.preprocess_image)

        #placeholder target tensor for creating models
        targets = None

    elif input_type == "tfrecord":
        #Create tensorflow iterators
        iterator = tfrecords.create_dataset(list_of_tfrecords, args.batch_size)
        next_element = iterator.get_next()

        #Split into inputs and targets
        inputs = next_element[0]
        targets = [next_element[1], next_element[2]]

        validation_generator = None

    else:
        raise ValueError(
            "{} input type is invalid. Only 'tfrecord' or 'for_generator' input types are accepted for model training"
            .format(input_type))

    # create the model
    if args.snapshot is not None:
        print('Loading model, this may take a second...')
        model = models.load_model(args.snapshot, backbone_name=args.backbone)
        training_model = model
        anchor_params = None
        if args.config and 'anchor_parameters' in args.config:
            anchor_params = parse_anchor_parameters(args.config)
        prediction_model = retinanet_bbox(model=model, anchor_params=anchor_params)
    else:
        weights = args.weights
        # default to imagenet if nothing else is specified
        if weights is None and args.imagenet_weights:
            weights = backbone.download_imagenet()

        print('Creating model, this may take a second...')
        if input_type == "fit_generator":
            num_of_classes = train_generator.num_classes()
        else:
            #Add background class
            num_of_classes = len(forest_object.labels.keys())

        model, training_model, prediction_model = create_models(
            backbone_retinanet=backbone.retinanet,
            num_classes=num_of_classes,
            weights=weights,
            multi_gpu=args.multi_gpu,
            freeze_backbone=args.freeze_backbone,
            lr=args.lr,
            config=args.config,
            targets=targets,
            freeze_layers=args.freeze_layers)

    # print model summary
    print(model.summary())

    # this lets the generator compute backbone layer shapes using the actual backbone model
    if 'vgg' in args.backbone or 'densenet' in args.backbone:
        train_generator.compute_shapes = make_shapes_callback(model)
        if validation_generator:
            validation_generator.compute_shapes = train_generator.compute_shapes

    # create the callbacks
    callbacks = create_callbacks(model, training_model, prediction_model,
                                 validation_generator, args, comet_experiment)

    if not args.compute_val_loss:
        validation_generator = None

    # start training
    if input_type == "fit_generator":
        history = training_model.fit_generator(generator=train_generator,
                                               steps_per_epoch=args.steps,
                                               epochs=args.epochs,
                                               verbose=1,
                                               callbacks=callbacks,
                                               workers=args.workers,
                                               use_multiprocessing=args.multiprocessing,
                                               max_queue_size=args.max_queue_size,
                                               validation_data=validation_generator)
    elif input_type == "tfrecord":

        #Fit model
        history = training_model.fit(x=inputs,
                                     steps_per_epoch=args.steps,
                                     epochs=args.epochs,
                                     callbacks=callbacks)
    else:
        raise ValueError(
            "{} input type is invalid. Only 'tfrecord' or 'for_generator' input types are accepted for model training"
            .format(input_type))

    #Assign history to deepforest model class
    forest_object.history = history

    #return trained model
    return model, prediction_model, training_model



from setuptools import setup, find_packages
import setuptools
from setuptools.extension import Extension
from distutils.command.build_ext import build_ext as DistUtilsBuildExt

NAME ='deepforest'
VERSION = '0.2.13'
DESCRIPTION = 'Tree crown prediction using deep learning retinanets'
URL = 'https://github.com/Weecology/DeepForest'
AUTHOR = 'Ben Weinstein'
LICENCE = 'MIT'
LONG_DESCRIPTION = """
# Deepforest

## Full documentation
[http://deepforest.readthedocs.io/en/latest/](http://deepforest.readthedocs.io/en/latest/)

## Installation

Compiled wheels have been made for linux, osx and windows

```
#Install DeepForest
pip install DeepForest
```


Or install the latest version from Github
```
git clone https://github.com/weecology/DeepForest.git

conda env create --file=environment.yml

conda activate DeepForest
```

## Get in touch
See the [GitHub Repo](https://github.com/Weecology/deepforest) to see the
source code or submit issues and feature requests.

## Citation
If you use this software in your research please cite it as:

Geographic Generalization in Airborne RGB Deep Learning Tree Detection
Ben. G. Weinstein, Sergio Marconi, Stephanie A. Bohlman, Alina Zare, Ethan P. White
bioRxiv 790071; doi: https://doi.org/10.1101/790071


## Acknowledgments
Development of this software was funded by
[the Gordon and Betty Moore Foundation's Data-Driven Discovery Initiative](http://www.moore.org/programs/science/data-driven-discovery) through
[Grant GBMF4563](http://www.moore.org/grants/list/GBMF4563) to Ethan P. White.
"""

#DeepForest wraps of fork of keras-retinanet, see https://github.com/fizyr/keras-retinanet/blob/master/setup.py

class BuildExtension(setuptools.Command):
    description     = DistUtilsBuildExt.description
    user_options    = DistUtilsBuildExt.user_options
    boolean_options = DistUtilsBuildExt.boolean_options
    help_options    = DistUtilsBuildExt.help_options

    def __init__(self, *args, **kwargs):
        from setuptools.command.build_ext import build_ext as SetupToolsBuildExt

        # Bypass __setatrr__ to avoid infinite recursion.
        self.__dict__['_command'] = SetupToolsBuildExt(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._command, name)

    def __setattr__(self, name, value):
        setattr(self._command, name, value)

    def initialize_options(self, *args, **kwargs):
        return self._command.initialize_options(*args, **kwargs)

    def finalize_options(self, *args, **kwargs):
        ret = self._command.finalize_options(*args, **kwargs)
        import numpy
        self.include_dirs.append(numpy.get_include())
        return ret

    def run(self, *args, **kwargs):
        return self._command.run(*args, **kwargs)


extensions = [
    Extension(
        'keras_retinanet.utils.compute_overlap',
        ['keras_retinanet/utils/compute_overlap.pyx']
    ),
]

setup(name=NAME,
      version=VERSION,
      python_requires='>=3',
      description=DESCRIPTION,
      long_description = LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      url=URL,
      author=AUTHOR,
      license=LICENCE,
      packages=find_packages(),
      include_package_data=True,
      cmdclass         = {'build_ext': BuildExtension},      
      install_requires=["keras > 2.3.0","keras-resnet==0.1.0","six","scipy","tensorflow==1.14.0","Pillow","pandas","opencv-python","pyyaml","slidingwindow","matplotlib","xmltodict","tqdm","progressbar2"],
      ext_modules    = extensions,
      setup_requires = ["cython>=0.28", "numpy>=1.14.0"],    
      zip_safe=False)



# -*- coding: utf-8 -*-
"""Top-level package for DeepForest."""
__author__ = """Ben Weinstein"""
__email__ = 'ben.weinstein@weecology.org'
__version__ = '0.2.13'

import os

_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_data(path):
    return os.path.join(_ROOT, 'data', path)



# test_utilities
from deepforest import utilities
from deepforest import get_data
import pytest
import os
import pandas as pd
import numpy as np
from deepforest import deepforest

@pytest.fixture()
def annotations():
    annotations = utilities.xml_to_annotations(get_data("OSBS_029.xml"))
    annotations_file = "tests/data/OSBS_029.csv"
    annotations.to_csv(annotations_file,index=False, header=False)
    return annotations_file

@pytest.fixture()
def config():
    config  = utilities.read_config(get_data("deepforest_config.yml"))
    return config
    
def test_xml_to_annotations():
    annotations = utilities.xml_to_annotations(xml_path = get_data("OSBS_029.xml"))
    print(annotations.shape)
    assert annotations.shape == (61 ,6)
    
    #bounding box extents should be int
    assert annotations["xmin"].dtype == np.int64
    
def test_create_classes(annotations):
    classes_file = utilities.create_classes(annotations_file=annotations)
    assert os.path.exists(classes_file)

def test_number_of_images(annotations):
    n = utilities.number_of_images(annotations_file=annotations)
    assert n == 1
    
def test_format_args(annotations, config):
    classes_file = utilities.create_classes(annotations)
    arg_list = utilities.format_args(annotations,classes_file, config)
    assert isinstance(arg_list, list)

def test_format_args_steps(annotations, config):
    classes_file = utilities.create_classes(annotations)    
    arg_list = utilities.format_args(annotations,classes_file, config, images_per_epoch=2)
    assert isinstance(arg_list, list)
    
    #A bit ugly, but since its a list, what is the argument after --steps to assert
    steps_position = np.where(["--steps" in x for x in arg_list])[0][0] + 1
    assert arg_list[steps_position] == '2'

def test_use_release():
    #Download latest model from github release
    release_tag, weights = utilities.use_release()
    assert os.path.exists(get_data("NEON.h5"))    